from simulate import Simulation
import numpy as np
import matplotlib.pyplot as plt

N = 2
dt = 1e-4
t = 0.0
tEnd = 1.0

pos  = np.random.uniform(size=(N,3))
vel = np.random.uniform(size=(N,3))
mass = np.ones(N)

tenbody = Simulation(N,dt,t,tEnd)
tenbody.initial_conditions(pos,vel,mass)
positions = tenbody.run()
tenbody.plot(three_dimensional=True)




from solvers import acceleration, leapfrog
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import animation


class Simulation():
    """
    Main simulation class for GravityLab

    Attributes
    ----------
    N    : Number of bodies
    dt   : Timestep
    t    : Start time
    tEnd : End time

    Methods
    -------
    initial_conditions : Initial conditions for simulation
    run                : Run simulation
    plot               : Plot positions

    """
    def __init__(self, N, dt, t, tEnd):
        """
        Initialize a simulation

        Parameters
        ----------
        N    : Number of bodies
        dt   : Timestep
        t    : Start time
        tEnd : End time

        """
        self.N = N
        self.dt = dt
        self.t = t
        self.tEnd = tEnd

    def initial_conditions(self, pos, vel, mass):
        """
        Initial conditions for a simulation

        Parameters
        ----------
        pos : N x 3 array of positions
        vel : N x 3 array of velocities
        mass : N x 1 array of masses

        """
        self.pos = pos
        self.vel = vel
        self.mass = mass
        self.init_pos = pos

    def run(self, verbose=False):
        """
        Run the simulation

        Parameters
        ----------
        verbose (optional) : Give more verbose output

        Returns
        -------
        pos_arr : N x 3 array of positions
        
        """
        Nt = int((self.tEnd - self.t)/self.dt)
        pos_lst = []

        acc = acceleration(self.pos, self.mass)

        for _ in range(Nt):
            self.pos, acc = leapfrog(self.pos, self.vel, acc, self.mass, self.dt)
            pos_lst.append(self.pos.copy())
        
        pos_arr = np.vstack(pos_lst)
        self.pos = pos_arr

        if verbose==True:
            print(pos_arr)

        return pos_arr

    def plot(self, three_dimensional=False, start_pos=True, color='blue', alpha = 0.6):
        """
        Plot positions of bodies

        Parameters
        ----------
        three_dimensional (optional) : 3D plotting
        color (optional)             : Desired colour of markers
        alpha (optional)             : Opacity of markers

        """
        if three_dimensional == True:
            fig = plt.figure()
            ax = plt.axes(projection='3d')
            if start_pos==True:
                ax.scatter3D(self.pos[:,0], self.pos[:,1], self.pos[:,2], color=color, alpha=alpha)
                ax.scatter3D(self.init_pos[0,0], self.init_pos[0,1], self.init_pos[0,2], color='red', alpha=alpha)
            else:
                ax.scatter3D(self.pos[:,0], self.pos[:,1], self.pos[:,2], color=color, alpha=alpha)
            plt.show()
        else:
            if start_pos==True:
                plt.scatter(self.pos[:,0], self.pos[:,1], color=color, alpha=alpha)
                plt.scatter(self.init_pos[0,0], self.init_pos[0,1], color='red')
            else:
                plt.scatter(self.pos[:,0], self.pos[:,1], color=color, alpha=alpha) 
            plt.show()
    
    def init_animation(self):
        """
        Initialize animation
        """
        fig, ax = plt.subplots()
        scatterplot, = ax.scatter([], [])
        scatterplot.set_data([], [])
        self.scatterplot = scatterplot
        return self.scatterplot,

    def animate_func(self, i):
        """
        Animate positions
        """
        acc = acceleration(self.pos, self.mass)
        self.pos, acc = leapfrog(self.pos, self.vel, acc, self.mass, self.dt)
        scatterplot = self.scatterplot
        for i in range(self.N):
            scatterplot.set_data(self.pos[:,0], self.pos[:,1])

        return self.scatterplot,
    
    def animate(self):
        """
        Animate
        """
        fig = plt.figure()
        Nt = int((self.tEnd - self.t)/self.dt)
        anim = animation.FuncAnimation(fig, self.animate_func, init_func=self.init_animation,
                               frames=Nt, interval=10, blit=False)
        plt.show()
        return

        
        



#cython: language_level=3

cimport numpy as np
import numpy as np


def separation(np.ndarray[np.float64_t, ndim=1] x1, np.ndarray[np.float64_t, ndim=1] x2):
    """
    Calculate separations

    Parameters
    ----------
    x1 : 1D array
        Position of body 1

    x2 : 1D array
        Position of body 2

    Returns
    -------
    dx : 1D array
        Separation vector

    r : float
        Magnitude of separation vector
    """
    cdef np.float64_t [:] dx # Separation vector (1D memoryview)
    cdef np.float64_t r # Separation vector magnitude

    # Calculate separation and make it an array
    dx = x2 - x1
    dx = np.asarray(dx, dtype="float64")

    # Calculate the magnitude of the separation 
    r = np.sqrt(np.dot(dx, dx))

    return r, np.asarray(dx, dtype="float64")


def acceleration(np.ndarray[np.float64_t, ndim=2] pos, np.ndarray[np.float64_t, ndim=1] mass):
    """
    Calculate accelerations

    Parameters
    ----------
    pos  : 2D array
           N x 3 array of positions

    vel  : 2D array
           N x 3 array of positions

    mass : 1D array
           N x 1 array of masses
    
    Returns
    -------
    acc : 2D array
          N x 3 array of accelerations
    """
    cdef int N = pos.shape[0]   # Number of bodies

    # Initialize N x 3 memoryview of accelerations
    cdef np.float64_t [:,:] acc = np.zeros((N,3), dtype="float64")

    # Separation vector magnitudes
    cdef np.float64_t r
    cdef np.float64_t r1
    cdef np.float64_t r2

    # Separation vectors
    cdef np.ndarray [np.float64_t, ndim=1] dx
    cdef np.ndarray [np.float64_t, ndim=1] dx1
    cdef np.ndarray [np.float64_t, ndim=1] dx2

    # Masses of body 1 and 2
    cdef np.float64_t m1
    cdef np.float64_t m2

    # Variable for indexing
    cdef int i

    # Calculate separation
    r, dx = separation(pos[0,:], pos[1,:]) # separation between 1 and 2

    # Get masses of body 1 and 2
    m1 = mass[0]
    m2 = mass[1]
    
    # Acceleration on body 1
    acc[0,0] =  m2 / r**3 * dx[0]
    acc[0,1] =  m2 / r**3 * dx[1]
    acc[0,2] =  m2 / r**3 * dx[2]

    # Acceleration on body 2
    acc[1,0] = -m1 / r**3 * dx[0]
    acc[1,1] = -m1 / r**3 * dx[1]
    acc[1,2] = -m1 / r**3 * dx[2]

    # Acceleration on other bodies due to bodies 1 and 2
    for i in range(2,N):
        r1, dx1 = separation(pos[i,:],pos[0,:]) # Separation between other bodies and body 1
        r2, dx2 = separation(pos[i,:],pos[1,:]) # Separation between other bodies and body 2

        # Calculate accelerations
        acc[i,0] = m1 / r1**3 * dx1[0] + m2 / r2**3 * dx2[0]
        acc[i,1] = m1 / r1**3 * dx1[1] + m2 / r2**3 * dx2[1]
        acc[i,2] = m1 / r1**3 * dx1[2] + m2 / r2**3 * dx2[2]

    return np.asarray(acc)


def euler_cromer(np.ndarray[np.float64_t, ndim=2] pos, np.ndarray[np.float64_t, ndim=2] vel, np.ndarray[np.float64_t, ndim=2] acc, double dt):
    """
    Integrate using the Euler-Cromer method

    v(t+1) = v(t) + a(t) x dt
    x(t+1) = x(t) + v(t+1) x dt

    Parameters
    ----------
    pos : 2D array
          N x 3 array of positions
    
    vel : 2D array
          N x 3 array of velocities

    acc : 2D Array
          N x 3 array of accelerations
    dt  : double
          Timestep

    Returns
    -------
    pos : 2D array
          N x 3 array of new positions
    """
    vel += acc * dt
    pos += vel * dt

    return pos



def leapfrog(np.ndarray[np.float64_t, ndim=2] pos, np.ndarray[np.float64_t, ndim=2] vel, np.ndarray[np.float64_t, ndim=2] acc, np.ndarray[np.float64_t, ndim=1] mass, double dt):
    """
    Integrate using the Leapfrog method.

    v(t+1/2) = v(t) + a(t) x dt/2
    x(t+1) = x(t) + v(t+1/2) x dt
    a(t+1) = (G * m/((dx^2 + dy^2 + dz^2)^(3/2))) * dx * x
    v(t+1) = v(t+1/2) + a(t+1) x dt/2

    Parameters
    ----------
    pos : 2D array
          N x 3 array of positions
    
    vel : 2D array
          N x 3 array of velocities

    mass : 1D Array
           N x 1 array of masses
    dt   : double
            Timestep

    Returns
    -------
    pos : 2D array
          N x 3 array of new positions
    acc : 2D array
          N x 3 array of accelerations
    """
    # v(t+1/2) = v(t) + a(t) x dt/2
    vel += acc * dt/2

    # x(t+1) = x(t) + v(t+1/2) x dt
    pos += vel * dt
    
    # a(t+1) = (G * m/((dx^2 + dy^2 + dz^2)^(3/2))) * dx * x
    acc = acceleration(pos, mass)

    # v(t+1) = v(t+1/2) + a(t+1) x dt/2
    vel += acc * dt/2

    return pos, acc



from setuptools import setup
from setuptools import Extension
from Cython.Distutils import build_ext
import numpy

extensions = [Extension("gravitylab.solvers", ["gravitylab/solvers.pyx"])]
cmdclass = {"build_ext" : build_ext}

setup(
    name="GravityLab",
    version="2.2.1",
    packages=["gravitylab"],
    license="MIT",
    author="Kushaal Kumar Pothula",
    author_email="kushaalkumar.astronomer@gmail.com",
    url = "https://github.com/Kushaalkumar-pothula/gravitylab",
    description="Fast astrophysical N-body simulator",
    long_description=open("README.md").read(),
    long_description_content_type='text/markdown',
    cmdclass=cmdclass,
    ext_modules=extensions,
    include_dirs=[numpy.get_include()],
    install_requires = [
        'numpy>=1.20.0',
        'matplotlib>=3.3.2'
    ]
)